# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

# Code from lib/canonical/lazr/pidfile.py modified to be used here.

__metaclass__ = type

import tempfile
import os
import time
import atexit
import sys
from signal import signal, SIGTERM

from tarmac.config import TarmacConfig


class PidfileExistsError(RuntimeError):
    def __init__(self, pidfile, pid, lifetime):
        self.pidfile = pidfile
        self.pid = pid
        self.lifetime = lifetime

    def __str__(self):
        return (
            'File %r reports that PID %r has existed for %d seconds' % (
                self.pidfile, self.pid, self.lifetime))

def make_pidfile():
    """Write the current process id to a PID file.

    Also installs an atexit handler to remove the file on process termination.

    Also installs a SIGTERM signal handler to remove the file on SIGTERM.
    If you install your own handler, you will want to call remove_pidfile
    inside it.
    """
    pidfile = TarmacConfig().PID_FILE
    fd, tempname = tempfile.mkstemp(dir=os.path.dirname(pidfile))
    outf = os.fdopen(fd, 'w')
    outf.write(str(os.getpid())+'\n')
    outf.flush()
    outf.close()
    # This introduces a race condition.  We're hoping it doesn't happen.
    if os.path.exists(pidfile):
        os.remove(tempname)
        raise PidfileExistsError(tempname, get_pid(), get_pidfile_lifetime())
    else:
        os.rename(tempname, pidfile)


# This could be used if the pidfile creation were not within a loop.
#
# def install_pidfile_removal():
#
#    atexit.register(remove_pidfile)
#    def remove_pidfile_handler(*ignored):
#        sys.exit(-1 * SIGTERM)
#    signal(SIGTERM, remove_pidfile_handler)


def remove_pidfile():
    """Remove the PID file.

    This should only be needed if you are overriding the default SIGTERM
    signal handler.
    """
    pidfile = TarmacConfig().PID_FILE
    pid = get_pid()
    if pid is None:
        return
    if pid == os.getpid():
        os.unlink(pidfile)


def get_pid():
    """Return the PID as an integer, or None

    May raise a ValueError if the PID file is corrupt.

    This method will only be needed by service or monitoring scripts.

    Currently no checking is done to ensure that the process is actually
    running, is healthy, or died horribly a while ago and its PID being
    used by something else. What we have is probably good enough.
    """
    pidfile = TarmacConfig().PID_FILE
    try:
        pid = open(pidfile).read()
        return int(pid)
    except IOError:
        return None
    except ValueError:
        raise ValueError("Invalid PID %r" % pid)


def get_pidfile_lifetime():
    """Return the number of seconds since the pidfile was created"""
    creation_time = os.stat(TarmacConfig().PID_FILE)[-1]
    now = time.time()
    return now - creation_time
