"""Tests for the BundleCommmand plug-in."""

from datetime import datetime, timedelta
import doctest
import os
import unittest

from testtools import TestCase, TestResult
from testtools.matchers import DocTestMatches

from tarmac.bin.registry import CommandRegistry
from tarmac.plugins.bundlecommand import (
    BundleCommand, TestResultFormatter, VerifyCommandFailed)
from tarmac.tests import BranchTestCase
from tarmac.tests.test_commands import FakeCommand
from tarmac.tests.mock import Thing, MockMergeProposal


class LoggingSMTPConnection(object):
    """An SMTPConnection double that logs sent email."""

    def __init__(self, log):
        self._log = log

    def send_email(self, message):
        self._log.append(message)


class Helpers:

    def make_formatter(self):
        proposal = Thing(
            source_branch=Thing(
                revision_count=123,
                display_name="lp:project/source"),
            target_branch=Thing(
                revision_count=120,
                display_name="lp:project"))
        start_time = datetime.utcnow()
        end_time = start_time + timedelta(hours=1)

        class SomeTest(TestCase):
            def test_ok(self):
                pass
            def test_fail(self):
                self.fail("oh no")
            def test_error(self):
                1/0
        self.fail_test = SomeTest('test_fail')
        self.error_test = SomeTest('test_error')
        test = unittest.TestSuite(
            [self.fail_test, self.error_test, SomeTest('test_ok')])
        result = TestResult()
        test.run(result)

        return TestResultFormatter(start_time, end_time, result, proposal)


class TestTestResultFormatter(TestCase, Helpers):
    """Tests for `TestResultFormatter`."""

    expected_failure_summary_email = ("""\
Tests started at approximately %(start_time)s
Source: %(source_branch)s r%(source_revno)s
Target: %(target_branch)s r%(target_revno)s
<BLANKLINE>
%(num_tests)s tests run in %(duration)s, %(num_failures)s failures, %(num_errors)s errors
<BLANKLINE>
Failing tests
-------------
  %(fail_id)s
<BLANKLINE>
Tests with errors
-----------------
  %(error_id)s
<BLANKLINE>
======================================================================
FAILURE: test_fail...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
======================================================================
ERROR: test_error...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
<BLANKLINE>
(See the attached file for the complete log)
""")

    expected_failure_summary = ("""\
Tests started at approximately %(start_time)s
Source: %(source_branch)s r%(source_revno)s
Target: %(target_branch)s r%(target_revno)s
<BLANKLINE>
%(num_tests)s tests run in %(duration)s, %(num_failures)s failures, %(num_errors)s errors
<BLANKLINE>
Failing tests
-------------
  %(fail_id)s
<BLANKLINE>
Tests with errors
-----------------
  %(error_id)s
<BLANKLINE>
======================================================================
FAILURE: test_fail...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
======================================================================
ERROR: test_error...
----------------------------------------------------------------------
Text attachment: traceback
------------
Traceback (most recent call last):
...
------------
<BLANKLINE>
""")

    def test_format_result_with_attached_log(self):
        formatter = self.make_formatter()
        summary = formatter.format_result(attached_log=True)
        data = {
            'start_time': formatter.start_time,
            'source_branch': formatter.source_branch,
            'source_revno': formatter.source_revno,
            'target_branch': formatter.target_branch,
            'target_revno': formatter.target_revno,
            'num_tests': formatter.num_tests,
            'duration': formatter.duration,
            'num_failures': formatter.num_failures,
            'num_errors': formatter.num_errors,
            'fail_id': self.fail_test.id(),
            'error_id': self.error_test.id()
            }
        self.assertThat(summary, DocTestMatches(
            self.expected_failure_summary_email % data, doctest.REPORT_NDIFF|
            doctest.ELLIPSIS))

    def test_format_result_without_attached_log(self):
        formatter = self.make_formatter()
        summary = formatter.format_result(attached_log=False)
        data = {
            'start_time': formatter.start_time,
            'source_branch': formatter.source_branch,
            'source_revno': formatter.source_revno,
            'target_branch': formatter.target_branch,
            'target_revno': formatter.target_revno,
            'num_tests': formatter.num_tests,
            'duration': formatter.duration,
            'num_failures': formatter.num_failures,
            'num_errors': formatter.num_errors,
            'fail_id': self.fail_test.id(),
            'error_id': self.error_test.id()
            }
        self.assertThat(summary, DocTestMatches(
            self.expected_failure_summary % data, doctest.REPORT_NDIFF|
            doctest.ELLIPSIS))


class BundleCommandTests(BranchTestCase, Helpers):
    """Test the BundleCommand."""

    def setUp(self):
        """Set up additional data we need for tests."""
        super(BundleCommandTests, self).setUp()
        # Use a set here because that's the object passed to BundleCommand.
        self.collected_proposals = set(
            [(Thing(),
            MockMergeProposal(
                Thing(
                    display_name='lp:project/source',
                    revision_count=123,
                    subscribers=[
                        Thing(preferred_email_address=Thing(
                            email='foo@example.com'))]),
                Thing(
                    display_name='lp:project',
                    revision_count=120),
                status="Approved",
                votes=[
                    Thing(reviewer=Thing(
                        preferred_email_address=Thing(
                        email='foo.bar@example.com')))])),
            (Thing(),
            MockMergeProposal(
                Thing(
                    display_name='lp:project/source2',
                    revision_count=124,
                    subscribers=[
                        Thing(preferred_email_address=Thing(
                            email='foo@example.com'))]),
                Thing(
                    display_name='lp:project',
                    revision_count=120),
                status="Approved",
                votes=[
                    Thing(reviewer=Thing(
                        preferred_email_address=Thing(
                        email='foo.bar@example.com')))]))]
            )
        # Patch the plugin smtp_connection so emails are never actually sent.
        email_log = []
        smtp_connection = LoggingSMTPConnection(email_log)
        self.plugin = BundleCommand(smtp_connection=smtp_connection)
        registry = CommandRegistry(config=self.config)
        self.command = FakeCommand(registry)

    def test_run(self):
        """Test that the plug-in run without errors."""
        target = Thing(config=Thing(
            bundle_verify_command="/bin/true",
            debug_log_location=self.TEST_ROOT),
            tree=Thing(abspath=os.path.abspath))
        self.plugin.run(
            command=self.command, target=target, restore_revno=123,
            collected_proposals=self.collected_proposals)

    def test_run_failure(self):
        """Test that a failure raises the correct exception."""
        # No email in the log.
        self.assertEqual(self.plugin._smtp_connection._log, [])
        # Proposal status are approved.
        proposal1, proposal2 = self.collected_proposals
        self.assertEqual(proposal1[1].queue_status, "Approved")
        self.assertEqual(proposal2[1].queue_status, "Approved")
        # Proposals have no comments.
        self.assertEqual(proposal1[1].all_comments, [])
        self.assertEqual(proposal2[1].all_comments, [])

        target = self.branch1
        # Include the config to run the bundle verify command.
        setattr(target.config, 'bundle_verify_command', '/bin/false')
        setattr(target.config, 'debug_log_location', self.TEST_ROOT)
        # Set a fake method to target, so we can make sure it's called when
        # the plugin fails to run.
        def fake_uncommit_and_revert(restore_revno):
            setattr(target, 'fake_revno', restore_revno)
        target.uncommit_and_revert = fake_uncommit_and_revert

        # Run the plugin.
        self.plugin.run(
            command=self.command, target=target, restore_revno=123,
            collected_proposals=self.collected_proposals)

        # There are two emails in the log. One for each proposal.
        self.plugin._smtp_connection._log.sort(
            key=lambda email: email.get('Subject'))
        self.assertTrue(len(self.plugin._smtp_connection._log) == 2)
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('Subject'),
            'Test results: lp:project/source2 => lp:project: FAILURE')
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('Subject'),
            'Test results: lp:project/source => lp:project: FAILURE')

        # Proposals statuses are updated.
        proposal1, proposal2 = self.collected_proposals
        self.assertEqual(proposal1[1].queue_status, "Needs review")
        self.assertEqual(proposal2[1].queue_status, "Needs review")
        # Proposals have comments.
        self.assertEqual(len(proposal1[1].all_comments), 1)
        self.assertEqual(len(proposal2[1].all_comments), 1)

        # Check the patched target.uncommit_and_revert() was run.
        self.assertEqual(target.fake_revno, 123)

    def test_debug_log(self):
        """Test logs are kept after plugin run."""
        current_log = os.path.join(self.TEST_ROOT, 'current.log')
        previous_log = os.path.join(self.TEST_ROOT, 'previous.log')
        self.assertFalse(os.path.exists(current_log))
        # Create a fake log file in current.log, so it can be moved to
        # previous.log
        log_file = open(current_log, 'w')
        log_file.write('Foobar')
        log_file.close()
        # Set up config for plugin.
        target = Thing(config=Thing(
            bundle_verify_command='/bin/true',
            debug_log_location=self.TEST_ROOT),
            tree=Thing(abspath=os.path.abspath))
        # Run.
        self.plugin.run(
            command=self.command, target=target, restore_revno=123,
            collected_proposals=self.collected_proposals)
        # Check we have current.log in place with no content since /bin/true
        # doesn't actually gives any output.
        self.assertTrue(os.path.exists(current_log))
        self.assertEqual('', open(current_log, 'r').read())
        # Check fake log was moved correctly to previous.log and the content
        # match.
        self.assertTrue(os.path.exists(previous_log))
        self.assertEqual('Foobar', open(previous_log, 'r').read())

    def test_send_report_email(self):
        """Test send_report_email()."""
        formatter = self.make_formatter()

        proposals = [
            proposal for source, proposal in self.collected_proposals]
        proposals.sort(
            key=lambda proposal: proposal.source_branch.display_name)
        emails = self.plugin.get_subscribers_emails(proposals[0])

        self.plugin.send_report_email(
            emails, formatter, 'some gzip content')
        self.assertTrue(len(self.plugin._smtp_connection._log) == 1)
        email = self.plugin._smtp_connection._log.pop()
        self.assertEqual(
            email.get('To'), 'foo@example.com, foo.bar@example.com')
        self.assertEqual(
            email.get('Subject'),
            "Test results: lp:project/source => lp:project: FAILURE")
